# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Handle marking of images for vimiv."""

import os

from gi.repository import GObject


class Mark(GObject.Object):
    """Handle marking of images.

    Attributes:
        marked: List of currently marked images.

        _app: The main vimiv application to interact with.
        _marked_bak: List of last marked images to be able to toggle marks.

    Signals:
        marks-changed: Emitted when new images were marked so other widgets can
            update their display.
    """

    def __init__(self, app):
        super(Mark, self).__init__()
        self._app = app
        self.marked = []
        self._marked_bak = []

    def mark(self):
        """Mark the current image."""
        # Check which image
        current = os.path.abspath(os.path.basename(self._app.get_pos(True)))
        # Toggle the mark
        if os.path.isfile(current):
            if current in self.marked:
                self.marked.remove(current)
            else:
                self.marked.append(current)
            self.emit("marks-changed", [current])
        else:
            self._app["statusbar"].message(
                "Marking directories is not supported", "warning")

    def toggle_mark(self):
        """Toggle mark status.

        If images are marked all marks are removed, otherwise the last marked
        images are re-marked.
        """
        if self.marked:
            self._marked_bak = self.marked
            self.marked = []
        else:
            self.marked, self._marked_bak = self._marked_bak, self.marked
        to_reload = self.marked + self._marked_bak
        self.emit("marks-changed", to_reload)

    def mark_all(self):
        """Mark all images."""
        # Get the correct filelist
        if self._app["library"].is_focus():
            files = []
            for fil in self._app["library"].files:
                files.append(os.path.abspath(fil))
        elif self._app.get_paths():
            files = self._app.get_paths()
        else:
            self._app["statusbar"].message("No image to mark", "error")
        # Add all to the marks
        for fil in files:
            if os.path.isfile(fil) and fil not in self.marked:
                self.marked.append(fil)
        self.emit("marks-changed", files)

    def mark_between(self):
        """Mark all images between the two last marks."""
        # Check if there are enough marks
        if len(self.marked) < 2:
            self._app["statusbar"].message("Not enough marks", "error")
            return
        start = self.marked[-2]
        end = self.marked[-1]
        # Get the correct filelist
        if self._app["library"].is_focus():
            files = []
            for fil in self._app["library"].files:
                if not os.path.isdir(fil):
                    files.append(os.path.abspath(fil))
        elif self._app.get_paths():
            files = self._app.get_paths()
        else:
            self._app["statusbar"].message("No image to mark", "error")
        # Find the images to mark
        for i, image in enumerate(files):
            if image == start:
                start = i
            elif image == end:
                end = i
        for i in range(start + 1, end):
            self.marked.insert(-1, files[i])
        self.emit("marks-changed", self.marked)


GObject.signal_new("marks-changed", Mark, GObject.SIGNAL_RUN_LAST, None,
                   (GObject.TYPE_PYOBJECT,))



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Main application class of vimiv."""

import os
import sys
import tempfile
from time import time

from gi.repository import Gdk, Gio, GLib, GObject, Gtk
from vimiv.commandline import CommandLine
from vimiv.completions import Completion
from vimiv.config_parser import parse_config
from vimiv.eventhandler import EventHandler
from vimiv.fileactions import ClipboardHandler, populate
from vimiv.information import Information
from vimiv.library import Library
from vimiv.log import Log
from vimiv.main_window import MainWindow
from vimiv.manipulate import Manipulate
from vimiv.mark import Mark
from vimiv.settings import settings
from vimiv.slideshow import Slideshow
from vimiv.statusbar import Statusbar
from vimiv.tags import TagHandler
from vimiv.transform import Transform
from vimiv.window import Window


class Vimiv(Gtk.Application):
    """Main vimiv application class inheriting from Gtk.Application.

    Attributes:
        debug: If True, write all messages and commands to log.

        _tmpdir: tmpfile.TemporaryDirectory used when running with
            --temp-basedir
        _widgets: Dictionary of vimiv widgets.
            widgets[widget-name] = Gtk.Widget
        _paths: List of paths for images.
        _index: Current position in paths.

    Signals:
        widget-layout-changed: Emitted when the layout of the widgets changed in
            some way. This allows other widgets to trigger an update when this
            happens, e.g. rezoom an image when the library was toggled.
        paths-changed: Emitted when the paths have or may have changed. This
            allows other widgets to reload their filelist and update any
            information accordingly.
    """

    def __init__(self, running_tests=False):
        """Create the Gtk.Application and connect the activate signal.

        Args:
            running_tests: If True, running from test suite. Do not show pop-up
                windows and do not parse user configuration files.
        """
        # Init application and set default values
        app_id = "org.vimiv" + str(time()).replace(".", "")
        super(Vimiv, self).__init__(application_id=app_id)
        self.set_flags(Gio.ApplicationFlags.HANDLES_OPEN)
        self.connect("activate", self.activate_vimiv)
        self._paths = []
        self._index = 0
        self._widgets = {}
        self.debug = False
        self._tmpdir = None
        self.running_tests = running_tests
        # Set up all commandline options
        self._init_commandline_options()

    def do_open(self, files, n_files, hint):
        """Open files by populating self._paths and self._index.

        Args:
            files: List of Gio.Files.
            n_files: Number of files.
            hint: Special string for user, always empty in vimiv.
        """
        filenames = [fil.get_path() for fil in files]
        # Move to the directory of the first argument
        if os.path.isdir(filenames[0]):
            working_directory = filenames[0]
        else:
            working_directory = os.path.dirname(filenames[0])
        if os.path.exists(working_directory):
            os.chdir(working_directory)
        # Populate list of images
        recursive = settings["recursive"].get_value()
        shuffle = settings["shuffle"].get_value()
        self.populate(filenames, recursive=recursive, shuffle_paths=shuffle)

        # Activate vimiv after opening files
        self.activate_vimiv(self)

    def do_handle_local_options(self, options):
        """Handle commandline arguments.

        Args:
            options: The dictionary containing all options given to the
                commandline.
        Return:
            Exit-code or -1 to continue
        """
        def set_option(name, setting, value=0):
            """Set a setting in settings according to commandline option.

            Args:
                name: Name of the commandline option
                setting: Name of setting in settings to modify.
                value: One of 0, 1 or 2.
                    0: Set setting to False.
                    1: Set setting to True.
                    2: Set setting to value given in the commandline.
            """
            if options.contains(name):
                valuedict = {0: "false", 1: "true"}
                if value == 2:
                    valuedict[2] = options.lookup_value(name).unpack()
                settings.override(setting, valuedict[value])

        # Show Gtk warnings if the debug option is given
        if options.contains("debug"):
            self.debug = True
        # If the version option is given, print version and exit 0
        if options.contains("version"):
            information = Information()
            print(information.get_version())
            return 0
        # Temp basedir removes all current settings and sets them to default
        if options.contains("temp-basedir"):
            self._tmpdir = tempfile.TemporaryDirectory(prefix="vimiv-")
            tmp = self._tmpdir.name
            # Export environment variables for thumbnails, tags, history and
            # logs
            os.environ["XDG_CACHE_HOME"] = os.path.join(tmp, "cache")
            os.environ["XDG_CONFIG_HOME"] = os.path.join(tmp, "config")
            os.environ["XDG_DATA_HOME"] = os.path.join(tmp, "data")
        # Create settings as soon as we know which config files to use
        elif options.contains("config"):
            configfile = options.lookup_value("config").unpack()
            parse_config(commandline_config=configfile,
                         running_tests=self.running_tests)
        else:
            parse_config(running_tests=self.running_tests)

        # If we start from desktop, move to the wanted directory
        # Else if the input does not come from a tty, e.g. find "" | vimiv, set
        # paths and index according to the input from the pipe
        if options.contains("start-from-desktop"):
            os.chdir(settings["desktop_start_dir"].get_value())
        elif not sys.stdin.isatty():
            try:
                tty_paths = []
                for line in sys.stdin:
                    tty_paths.append(line.rstrip("\n"))
                self.populate(tty_paths)
            except TypeError:
                pass  # DebugConsoleStdIn is not iterable

        set_option("bar", "display_bar", 1)
        set_option("no-bar", "display_bar", 0)
        set_option("library", "start_show_library", 1)
        set_option("no-library", "start_show_library", 0)
        set_option("fullscreen", "start_fullscreen", 1)
        set_option("no-fullscreen", "start_fullscreen", 0)
        set_option("shuffle", "shuffle", 1)
        set_option("no-shuffle", "shuffle", 0)
        set_option("recursive", "recursive", 1)
        set_option("no-recursive", "recursive", 0)
        set_option("slideshow", "start_slideshow", 1)
        set_option("slideshow-delay", "slideshow_delay", 2)
        set_option("geometry", "geometry", 2)

        return -1  # To continue

    def activate_vimiv(self, app):
        """Starting point for the vimiv application.

        Args:
            app: The application itself.
        """
        self._init_widgets()
        self._create_window_structure()
        app.add_window(self["window"])
        # Show everything and then hide whatever needs to be hidden
        self["window"].show_all()
        self["manipulate"].hide()
        self["commandline"].hide()
        self["completions"].hide()
        # Statusbar depending on setting
        if not settings["display_bar"].get_value():
            self["statusbar"].hide()
        self["statusbar"].set_separator_height()
        # Try to generate imagelist recursively from the current directory if
        # recursive is given and no paths exist
        if settings["recursive"].get_value() and not self._paths:
            shuffle = settings["shuffle"].get_value()
            self.populate([os.getcwd()], True, shuffle)
        # Show the image if an imagelist exists
        if self._paths:
            self["image"].load()
            # Show library at the beginning?
            if not settings["start_show_library"].get_value():
                self["library"].grid.hide()
            self["main_window"].grab_focus()
            # Start in slideshow mode?
            if settings["start_slideshow"].get_value():
                self["slideshow"].toggle()
        else:
            # Slideshow without paths makes no sense
            self["slideshow"].running = False
            self["library"].focus()
            if settings["expand_lib"].get_value():
                self["main_window"].hide()

    def quit_wrapper(self, force=False):
        """Quit the applications, print marked files and save history.

        Args:
            force: If True quit even if an image was edited.
        """
        # Check for running external processes
        if self["commandline"].running_processes:
            if force:
                for p in self["commandline"].running_processes:
                    p.kill()
            else:
                processes = [p.args
                             for p in self["commandline"].running_processes]
                message = "Running external processes: %s. Add ! to force." \
                          % (", ".join(processes))
                self["statusbar"].message(message, "warning")
                return
        # Check if image has been edited
        if self["manipulate"].check_for_edit(force):
            return
        for image in self["mark"].marked:
            print(image)
        # Run remaining rotate and flip threads
        self["transform"].apply()
        # Save the history
        self["commandline"].write_history()
        # Write to log
        self["log"].write_message("Exited", "time")
        # Cleanup tmpdir
        if self._tmpdir:
            self._tmpdir.cleanup()
        # Call Gtk.Application.quit()
        self.quit()

    def get_focused_widget(self):
        """Return the currently focused widget as string."""
        if self["library"].is_focus():
            return "lib"
        elif self["thumbnail"].is_focus():
            return "thu"
        elif self["manipulate"].is_visible():
            return "man"
        return "im"

    def get_pos(self, get_filename=False):
        """Get the current position in the focused widget.

        get_filename: If True return filename instead of position.

        Return:
            Current position as number or filename.
        """
        focused_widget = self.get_focused_widget()

        # Get position and name
        filelist = self._paths
        position = self._index
        if focused_widget == "lib":
            position = self["library"].get_position()
            filelist = self["library"].files
        elif focused_widget == "thu":
            position = self["thumbnail"].get_position()

        if get_filename:
            return filelist[position] if position < len(filelist) else ""
        return position

    def get_paths(self):
        return self._paths

    def get_index(self):
        return self._index

    def get_path(self):
        return self._paths[self._index]

    def update_index(self, diff):
        self._index = (self._index + diff) % len(self._paths)

    def remove_path(self, path):
        self._paths.remove(path)

    def populate(self, args, recursive=False, shuffle_paths=False,
                 expand_single=True):
        """Simple wrapper for fileactions.populate.

        Args:
            args: Paths given.
            recursive: If True search path recursively for images.
            shuffle_paths: If True shuffle found paths randomly.
            expand_single: If True, populate a complete filelist with images
                from the same directory as the single argument given.
        """
        self._paths, self._index = populate(
            args, recursive=recursive, shuffle_paths=shuffle_paths,
            expand_single=expand_single)

    def _init_commandline_options(self):
        """Add all possible commandline options."""
        def add_option(name, short, help_str, flags=GLib.OptionFlags.NONE,
                       arg=GLib.OptionArg.NONE, value=None):
            """Add a single option.

            Args:
                name: Long name of the option, called via --name.
                short: Short name of the option, called via -s.
                help_str: Displayed help text when calling --help.
                flags: GLib options.
                arg: Type of the value to give after option.
                value: Name of the value to give after the option.
            """
            if short:
                short = ord(short)
            self.add_main_option(name, short, flags, arg, help_str, value)
        add_option("bar", "b", "Display statusbar")
        add_option("no-bar", "B", "Hide statusbar")
        add_option("fullscreen", "f", "Start fullscreen")
        add_option("no-fullscreen", "F", "Do not start fullscreen")
        add_option("library", "l", "Display library")
        add_option("no-library", "L", "Hide library")
        add_option("recursive", "r", "Search directory recursively for images")
        add_option("no-recursive", "R",
                   "Do not search directory recursively for images")
        add_option("shuffle", "s", "Shuffle filelist")
        add_option("no-shuffle", "S", "Do not shuffle filelist")
        add_option("version", "v", "Print version information and exit")
        add_option("start-from-desktop", 0,
                   "Start using the desktop_start_dir as path")
        add_option("slideshow", 0, "Start slideshow at startup")
        add_option("slideshow-delay", 0, "Set slideshow delay",
                   arg=GLib.OptionArg.DOUBLE, value="delay")
        add_option("geometry", "g", "Set the starting geometry",
                   arg=GLib.OptionArg.STRING, value="GEOMETRY")
        add_option("temp-basedir", 0, "Use a temporary basedir")
        add_option("config", 0, "Use FILE as local configuration file",
                   arg=GLib.OptionArg.STRING, value="FILE")
        add_option("debug", 0, "Run in debug mode")

    def _init_widgets(self):
        """Create all the other widgets and add them to the class."""
        self["eventhandler"] = EventHandler(self)
        self["transform"] = Transform(self)
        self["commandline"] = CommandLine(self)
        self["tags"] = TagHandler(self)
        self["mark"] = Mark(self)
        self["clipboard"] = ClipboardHandler(self)
        self["statusbar"] = Statusbar(self)
        self["completions"] = Completion(self)
        self["slideshow"] = Slideshow(self)
        self["main_window"] = MainWindow(self)
        self["image"] = self["main_window"].image
        self["thumbnail"] = self["main_window"].thumbnail
        self["library"] = Library(self)
        self["manipulate"] = Manipulate(self)
        self["information"] = Information()
        self["window"] = Window(self)
        # Generate completions as soon as commands exist
        self["commandline"].init_commands()
        self["completions"].generate_commandlist()
        self["log"] = Log(self)

    def _create_window_structure(self):
        """Generate the structure of all the widgets and add it to the window.

        There is a large Gtk.Grid() to organize the widgets packed into a
        Gtk.Overlay(). The commandline is added as overlay.
        """
        main_grid = Gtk.Grid()
        main_grid.attach(self["library"].grid, 0, 0, 1, 1)
        main_grid.attach(self["main_window"], 1, 0, 1, 1)
        main_grid.attach(self["manipulate"], 0, 1, 2, 1)
        main_grid.attach(self["statusbar"].separator, 0, 2, 2, 1)

        overlay_grid = Gtk.Grid()
        overlay_grid.attach(self["statusbar"], 0, 0, 1, 1)
        overlay_grid.attach(self["completions"].info, 0, 1, 1, 1)
        overlay_grid.attach(self["commandline"], 0, 2, 1, 1)
        overlay_grid.set_valign(Gtk.Align.END)

        # Make it nice using CSS
        overlay_grid.set_name("OverLay")
        style = Gtk.Window().get_style_context()
        bg = style.get_background_color(Gtk.StateType.NORMAL)
        color_str = "#OverLay { background: " + bg.to_string() + "; }"
        command_provider = Gtk.CssProvider()
        command_css = color_str.encode()
        command_provider.load_from_data(command_css)
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(), command_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        # Overlay contains grid mainly and adds commandline as floating
        overlay = Gtk.Overlay()
        overlay.add(main_grid)
        overlay.add_overlay(overlay_grid)
        self["window"].add(overlay)

    def __getitem__(self, name):
        """Convenience method to access widgets via self[name].

        Args:
            name: Name of the widget to access.
        Return:
            The widget if it exists. None else.
        """
        return self._widgets[name]

    def __setitem__(self, name, widget):
        """Convenience method to access widgets via self[name].

        Args:
            name: Name of the widget to add.
            item: The actual widget to which name will refer.
        """
        self._widgets[name] = widget


# Initiate signals for the Vimiv class
GObject.signal_new("widget-layout-changed", Vimiv, GObject.SIGNAL_RUN_LAST,
                   None, (GObject.TYPE_PYOBJECT,))
GObject.signal_new("paths-changed", Vimiv, GObject.SIGNAL_RUN_LAST,
                   None, (GObject.TYPE_PYOBJECT,))



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Provides a class to manage settings and methods dealing with conversion."""

import os

from gi.repository import GObject
from vimiv.exceptions import (NotABoolean, NotANumber, SettingNotFoundError,
                              StringConversionError)
from vimiv.helpers import get_boolean, get_float, get_int


class Setting(object):
    """Stores a setting and its attributes.

    This class should not be used directly. Instead it is used as BaseClass for
    different types of settings.

    Attributes:
        name: Name of the setting as string.

        _datatype: Python type of the setting.
        _default_value: Default value of the setting stored in its python type.
        _value: Value of the setting stored in its python type.
    """

    def __init__(self, name, default_value):
        """Initialize attributes with default values.

        Args:
            name: Name of the setting to initialize.
            default_value: Default value of the setting to start with.
        """
        self.name = name
        self._default_value = default_value
        self._value = default_value
        self._n = 0

    def get_default(self):
        return self._default_value

    def get_value(self):
        return self._value

    def is_default(self):
        return self._value == self._default_value

    def set_to_default(self):
        self._value = self._default_value


class BoolSetting(Setting):
    """Stores a boolean setting."""

    def override(self, new_value):
        self._value = get_boolean(new_value)

    def toggle(self):
        self._value = not self._value


class IntSetting(Setting):
    """Stores an integer setting."""

    def override(self, new_value):
        self._value = get_int(new_value)

    def add(self, value, multiplier):
        """Add a value to the currently stored integer.

        Args:
            value: The integer value to add.
            multiplier: Integer to multiply value with.
        """
        value = get_int(value, allow_sign=True)
        multiplier = get_int(multiplier, allow_sign=False)
        self._value += value * multiplier


class FloatSetting(Setting):
    """Stores a float setting."""

    def override(self, new_value):
        self._value = get_float(new_value)

    def add(self, value, multiplier):
        """Add a value to the currently stored float.

        Args:
            value: The float value to add.
            multiplier: Float to multiply value with.
        """
        value = get_float(value, allow_sign=True)
        multiplier = get_float(multiplier, allow_sign=False)
        self._value += value * multiplier


class ThumbnailSizeSetting(Setting):
    """Stores a thumbnail size setting.

    This setting is stored as a python tuple with two equal integer values. The
    integer value must be one of 64, 128, 256, 512.
    """

    def override(self, new_value):
        """Override the setting with a new thumbnail size.

        Args:
            new_value: String in the form of "(val, val)" for the new value.
        """
        new_value = new_value.lstrip("(").rstrip(")")
        int_tuple = new_value.split(",")
        if len(int_tuple) != 2:
            error = "Tuple must be of the form (val, val)"
            raise StringConversionError(error)
        int_tuple = (get_int(int_tuple[0]), get_int(int_tuple[1]))
        if int_tuple[0] != int_tuple[1]:
            error = "Thumbnail width and height must be equal"
            raise StringConversionError(error)
        if int_tuple[0] not in [64, 128, 256, 512]:
            error = "Thumbnail size must be one of 64, 128, 256 and 512"
            raise StringConversionError(error)
        self._value = tuple(int_tuple)


class GeometrySetting(Setting):
    """Stores a geometry setting.

    This setting is stored as python tuple of the form (width, height) and can
    be set from a string in the form "WIDTHxHEIGHT".
    """

    def override(self, new_value):
        """Override the setting with a new geometry.

        Args:
            new_value: String in the form of "WIDTHxHEIGHT" for the new value.
        """
        geometry = new_value.split("x")
        if len(geometry) != 2:
            error = "Geometry must be of the form WIDTHxHEIGHT"
            raise StringConversionError(error)
        self._value = (get_int(geometry[0]), get_int(geometry[1]))


class DirectorySetting(Setting):
    """Stores a directory setting.

    This setting gets stored as a python string containing the absolute path of
    the directory. Override checks if the directory exists.
    """

    def override(self, new_value):
        directory = os.path.abspath(os.path.expanduser(new_value))
        if os.path.isdir(directory):
            self._value = directory
        else:
            error = "Directory '%s' does not exist" % (new_value)
            raise StringConversionError(error)


class MarkupSetting(Setting):
    """Stores a markup setting to highlight e.g. search results.

    This setting gets stored as a python string but it must be closeable via
    "</span>". The valid default is '<span foreground="#875FFF">'.
    """

    def override(self, new_value):
        """Override the currently stored value with a new one performing checks.

        Args:
            new_value: The value used to override the current value.
        """
        new_value = new_value.strip()
        if not new_value.startswith("<span"):
            error = 'Markup must start with "<span"'
            raise StringConversionError(error)
        elif not new_value.endswith(">"):
            error = 'Markup must end with ">"'
            raise StringConversionError(error)
        self._value = new_value

    def surround(self, string):
        """Surround a string with the markup tags.

        Args:
            string: The string to surround.
        Return:
            self._value + string + "</span>"
        """
        return self._value + string + "</span>"


class SettingStorage(GObject.Object):
    """Stores all settings for vimiv.

    Settings can be accessed via SettingStorage["setting_name"] and changed by
    calling SettingStorage.override("setting_name", new_value).

    Attributes:
        _aliases: Dictionary containing all aliases.
        _settings: List of all Setting objects. Should not be accessed directly.
        _n: Integer for iterator.

    Signals:
        changed: Emitted when a setting was changed.
    """

    def __init__(self):
        super(SettingStorage, self).__init__()
        self._aliases = {}
        self._settings = [
            BoolSetting("start_fullscreen", False),
            BoolSetting("start_slideshow", False),
            FloatSetting("slideshow_delay", 2.0),
            BoolSetting("shuffle", False),
            BoolSetting("display_bar", True),
            ThumbnailSizeSetting("default_thumbsize", (128, 128)),
            GeometrySetting("geometry", (800, 600)),
            BoolSetting("search_case_sensitive", True),
            BoolSetting("incsearch", True),
            BoolSetting("recursive", False),
            BoolSetting("rescale_svg", True),
            FloatSetting("overzoom", 1.0),
            BoolSetting("copy_to_primary", False),
            IntSetting("commandline_padding", 6),
            IntSetting("thumb_padding", 10),
            IntSetting("completion_height", 200),
            BoolSetting("play_animations", True),
            BoolSetting("start_show_library", False),
            IntSetting("library_width", 300),
            BoolSetting("expand_lib", True),
            IntSetting("border_width", 0),
            MarkupSetting("markup", '<span foreground="#875FFF">'),
            BoolSetting("show_hidden", False),
            DirectorySetting("desktop_start_dir", os.path.expanduser("~")),
            IntSetting("file_check_amount", 30),
            BoolSetting("tilde_in_statusbar", True),
            BoolSetting("autosave_images", True)]
        self._n = 0

    def override(self, name, new_value=None):
        """Override a setting in the storage.

        Args:
            name: Name of the setting to override.
            new_value: Value to override the setting with. If None use default.
        """
        setting = self[name]
        if new_value is None:
            setting.set_to_default()
        else:
            setting.override(new_value)
        self.emit("changed", name)

    def toggle(self, name):
        """Toggle a setting in the storage.

        If the setting is not a boolean, an exception is raised.

        Args:
            name: Name of the setting to toggle.
        """
        if isinstance(self[name], BoolSetting):
            self[name].toggle()
            self.emit("changed", name)
        else:
            error = "Cannot toggle non boolean setting '%s'" % (name)
            raise NotABoolean(error)

    def add_to(self, name, value, multiplier):
        """Add a number to a setting in the storage.

        If the setting is not a number, an exception is raised.

        Args:
            name: Name of the setting to change.
            value: Value to add.
            multiplier: Multiply value by this. Used for num_str.
        """
        if isinstance(self[name], (IntSetting, FloatSetting)):
            self[name].add(value, multiplier)
            self.emit("changed", name)
        else:
            error = "Cannot add to non number setting '%s'" % (name)
            raise NotANumber(error)

    def reset(self):
        """Reset all settings to their default value."""
        for setting in self:
            setting.set_to_default()

    def set_aliases(self, aliases):
        self._aliases = aliases

    def get_aliases(self):
        return self._aliases

    def __getitem__(self, name):
        """Receive a setting via its name."""
        for setting in self._settings:
            if setting.name == name:
                return setting
        raise SettingNotFoundError("No setting called '%s'" % (name))

    def __iter__(self):
        self._n = 0
        return self

    def __next__(self):
        """Iterate over settings."""
        if (self._n - 1) < len(self._settings) - 1:
            self._n += 1
            return self._settings[self._n - 1]
        else:
            raise StopIteration


# Initiate signals for the SettingsStorage
GObject.signal_new("changed", SettingStorage, GObject.SIGNAL_RUN_LAST,
                   None, (GObject.TYPE_PYOBJECT,))


# Initialize an actual SettingsStorage object to work with
settings = SettingStorage()



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Manipulate part for vimiv."""

import os

from gi.repository import GdkPixbuf, Gtk
from vimiv import image_enhance
from vimiv.exceptions import StringConversionError
from vimiv.fileactions import edit_supported
from vimiv.helpers import get_int
from vimiv.imageactions import save_pixbuf


class Manipulate(Gtk.ScrolledWindow):
    """Manipulate class for vimiv.

    Includes the manipulation toolbar and all the actions that apply to it.

    Attributes:
        sliders: Dictionary containing bri, con and sat sliders.

        _app: The main vimiv application to interact with.
        _manipulations: Dictionary of possible manipulations. Includes
            brightness, contrast and saturation.
        _pixbuf: Full sized GdkPixbuf displayed in Image.
    """

    def __init__(self, app):
        """Create the necessary objects and settings.

        Args:
            app: The main vimiv application to interact with.
        """
        super(Manipulate, self).__init__()
        self._app = app

        # Defaults
        self._manipulations = {"bri": 0, "con": 0, "sat": 0}
        self._pixbuf = GdkPixbuf.Pixbuf()

        # A scrollable window so all tools are always accessible
        self.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.NEVER)
        # A grid in which everything gets packed
        grid = Gtk.Grid()
        grid.set_column_spacing(6)
        grid.set_border_width(6)
        grid.connect("key_press_event", self._app["eventhandler"].on_key_press,
                     "MANIPULATE")
        self.add(grid)

        # Sliders
        bri_slider = Gtk.Scale()
        bri_slider.connect("value-changed", self._set_slider_value, "bri")
        con_slider = Gtk.Scale()
        con_slider.connect("value-changed", self._set_slider_value, "con")
        sat_slider = Gtk.Scale()
        sat_slider.connect("value-changed", self._set_slider_value, "sat")
        self.sliders = {"bri": bri_slider, "con": con_slider, "sat": sat_slider}

        # Set some properties
        for slider in self.sliders.values():
            slider.set_range(-127, 127)
            slider.set_size_request(120, 20)
            slider.set_value(0)
            slider.set_digits(0)

        # Labels
        bri_label = Gtk.Label()
        bri_label.set_markup("\n<b>Bri</b>")
        con_label = Gtk.Label()
        con_label.set_markup("\n<b>Con</b>")
        sat_label = Gtk.Label()
        sat_label.set_markup("\n<b>Sat</b>")

        # Dummy grid as separator
        separator = Gtk.Grid()
        separator.set_hexpand(True)

        # Buttons
        button_yes = Gtk.Button(label="Accept")
        button_yes.connect("clicked", self._on_button_clicked, True)
        button_yes.set_size_request(80, 20)
        button_no = Gtk.Button(label="Cancel")
        button_no.connect("clicked", self._on_button_clicked, False)
        button_no.set_size_request(80, 20)

        # Pack everything into the grid
        for item in [bri_label, bri_slider, con_label, con_slider,
                     sat_label, sat_slider, separator, button_yes,
                     button_no]:
            grid.add(item)

    def check_for_edit(self, force):
        """Check if an image was edited.

        Args:
            force: If True move and ignore any editing.
        Return:
            0 if no edits were done or force, 1 else.
        """
        if force:
            self._manipulations = {"bri": 0, "con": 0, "sat": 0}
            return 0
        elif self._manipulations != {"bri": 0, "con": 0, "sat": 0}:
            self._app["statusbar"].message(
                "Image has been edited, add ! to force", "warning")
            return 1
        return 0

    def toggle(self):
        """Toggle the manipulation bar."""
        if self.is_visible():
            self.hide()
            self._app["main_window"].grab_focus()
            self._app["statusbar"].update_info()
        elif self._app.get_paths() and \
                self._app.get_focused_widget() not in ["lib", "thu"]:
            if os.path.islink(self._app.get_path()):
                self._app["statusbar"].message(
                    "Manipulating symbolic links is not supported", "warning")
            elif not edit_supported(self._app.get_path()):
                self._app["statusbar"].message(
                    "This filetype is not supported", "warning")
            else:
                self.show()
                self._pixbuf = self._app["image"].get_pixbuf_original()
                self.sliders["bri"].grab_focus()
                self._app["statusbar"].update_info()
        else:
            if self._app["thumbnail"].toggled:
                self._app["statusbar"].message(
                    "Manipulate not supported in thumbnail mode", "warning")
            elif self._app["library"].is_focus():
                self._app["statusbar"].message(
                    "Manipulate not supported in library", "warning")
            else:
                self._app["statusbar"].message("No image open to edit",
                                               "warning")

    def _apply(self, real=False):
        """Apply manipulations to image.

        Manipulations are the three sliders for brightness, contrast and
        saturation. They are applied to a thumbnail by default and can act on
        the real image.

        Args:
            real: If True, apply manipulations to the real image and save it.
        """
        pixbuf = self._pixbuf if real else self._pixbuf.copy()
        # Apply brightness, contrast and saturation
        pixbuf = image_enhance.enhance_bc(pixbuf, self._manipulations["bri"],
                                          self._manipulations["con"])
        pixbuf.saturate_and_pixelate(pixbuf, self._manipulations["sat"] + 1,
                                     False)
        # Show the edited pixbuf
        self._app["image"].set_pixbuf(pixbuf)
        # Save file if needed
        if real:
            save_pixbuf(pixbuf, self._app.get_path())

    def _set_slider_value(self, slider, name):
        """Set value of self._manipulations according to slider value.

        Args:
            slider: Gtk.Scale on which to operate.
            name: Name of slider to operate on.
        """
        val = slider.get_value()
        val /= 127
        # Change brightness, contrast or saturation
        self._manipulations[name] = val
        # Run the manipulation function
        self._apply()

    def focus_slider(self, name):
        """Set focus on one of the three sliders.

        Args:
            name: Name of slider to focus.
        """
        # If manipulate is not toggled, this makes no sense
        if not self.is_visible():
            self._app["statusbar"].message(
                "Focusing a slider only makes sense in manipulate", "error")
        elif name not in self.sliders:
            self._app["statusbar"].message(
                "No slider called " + name, "error")
        else:
            self.sliders[name].grab_focus()

    def change_slider(self, step="1"):
        """Change the value of the currently focused slider.

        Args:
            step: Step to edit the slider by.
        """
        try:
            step = get_int(step, allow_sign=True)
        except StringConversionError as e:
            self._app["statusbar"].message(str(e), "error")
            return
        for slider in self.sliders.values():
            if slider.is_focus():
                val = slider.get_value()
                step = self._app["eventhandler"].num_receive() * step
                val += step
                slider.set_value(val)

    def _on_button_clicked(self, button, accept=False):
        self.finish(accept)

    def finish(self, accept=False):
        """Finish manipulate mode.

        Args:
            accept: If True apply changes to image file. Else discard them.
        """
        # If manipulate is not toggled, this makes no sense
        if not self.is_visible():
            self._app["statusbar"].message(
                "Finishing manipulate only makes sense in manipulate", "error")
            return
        # Apply changes
        if accept:
            self._apply(real=True)
        # Reset all the manipulations
        self._manipulations = {"bri": 0, "con": 0, "sat": 0}
        for slider in self.sliders.values():
            slider.set_value(0)
        # Show the original image
        self._app["image"].load()
        # Done
        self.toggle()

    def cmd_edit(self, manipulation, num="0"):
        """Run the specified manipulation.

        Args:
            manipulation: Name of manipulation to run.
            num: Value for setting the slider.
        """
        # Catch buggy numbers
        if not num.isdigit():
            self._app["statusbar"].message(
                "Argument must be of type integer", "error")
            return
        if not self.is_visible():
            if not self._app.get_paths():
                self._app["statusbar"].message(
                    "No image to manipulate", "error")
                return
            else:
                self.toggle()
        self.focus_slider(manipulation)
        self.sliders[manipulation].set_value(int(num))



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Thumbnail part of vimiv."""

import os
from math import floor

from gi.repository import GdkPixbuf, GLib, Gtk
from vimiv.settings import settings
from vimiv.thumbnail_manager import ThumbnailManager


class Thumbnail(Gtk.IconView):
    """Thumbnail class for vimiv.

    Includes the iconview with the thumbnails and all actions that apply to it.

    Attributes:
        toggled: If True, thumbnail mode is open.

        _app: The main vimiv application to interact with.
        _last_focused: Widget that was focused before thumbnail.
        _liststore: Gtk.ListStore containing thumbnail pixbufs and names.
        _markup: Markup string used to highlight search results.
        _thumbnail_manager: ThumbnailManager class to create and receive
            thumbnail files.
        _timer_id: ID of the currently running GLib.Timeout.
        _zoom_levels: List of tuples containing the possible thumbnail sizes.
        _zoom_level_index: Position in the possible_sizes list.
    """

    def __init__(self, app):
        """Create the necessary objects and settings.

        Args:
            app: The main application class to interact with.
        """
        super(Thumbnail, self).__init__()
        self._app = app

        # Settings
        self.toggled = False
        padding = settings["thumb_padding"].get_value()
        self._timer_id = GLib.Timeout
        self._markup = settings["markup"].get_value().replace("fore", "back")

        zoom_level = settings["default_thumbsize"].get_value()
        self._zoom_levels = [(64, 64), (128, 128), (256, 256), (512, 512)]
        self._zoom_level_index = self._zoom_levels.index(zoom_level)

        # Create the liststore and iconview
        self._liststore = Gtk.ListStore(GdkPixbuf.Pixbuf, str)
        self.connect("item-activated", self._on_activated)
        self.connect("key_press_event",
                     self._app["eventhandler"].on_key_press, "THUMBNAIL")
        self.connect("button_press_event",
                     self._app["eventhandler"].on_click, "THUMBNAIL")
        self.set_model(self._liststore)
        self.set_pixbuf_column(0)
        self.set_markup_column(1)

        self.set_item_width(0)
        self.set_item_padding(padding)
        self.last_focused = ""
        self._thumbnail_manager = ThumbnailManager()

        # Signals
        self._app["mark"].connect("marks-changed", self._on_marks_changed)
        self._app["transform"].connect("applied-to-file",
                                       self._on_transformations_applied_to_file)
        self._app["commandline"].search.connect("search-completed",
                                                self._on_search_completed)

    def _on_activated(self, iconview, path):
        """Select and show image when thumbnail was activated.

        Args:
            iconview: Gtk.IconView that emitted the signal.
            path: Gtk.TreePath of the activated thumbnail.
        """
        self.toggle(True)
        count = path.get_indices()[0] + 1
        self._app["eventhandler"].num_clear()
        self._app["eventhandler"].set_num_str(count)
        self._app["image"].move_pos()

    def toggle(self, select_image=False):
        """Toggle thumbnail mode.

        Args:
            select_image: If True an image was selected. Never focus the library
                then.
        """
        # Close
        if self.toggled:
            self._app["main_window"].switch_to_child(self._app["image"])
            if self.last_focused == "im" or select_image:
                self._app["main_window"].grab_focus()
            elif self.last_focused == "lib":
                self._app["library"].focus()
                # Re-expand the library if there is no image and the setting
                # applies
                if settings["expand_lib"].get_value() and \
                        not self._app["image"].get_pixbuf():
                    self._app["main_window"].hide()
                    self._app["library"].set_hexpand(True)
            self.toggled = False
        # Open thumbnail mode differently depending on where we come from
        elif self._app.get_paths() and self._app.get_focused_widget() == "im":
            self.last_focused = "im"
            self._show()
        elif self._app.get_focused_widget() == "lib":
            self.last_focused = "lib"
            self._app.populate(self._app["library"].files)
            if self._app.get_paths():
                self._app["library"].set_hexpand(False)
                self._app["main_window"].show()
                self._show()
            else:
                self._app["statusbar"].message(
                    "No images in directory", "error")
                return
        else:
            self._app["statusbar"].message("No open image", "error")
            return
        # Manipulate bar is useless in thumbnail mode
        if self._app["manipulate"].is_visible():
            self._app["manipulate"].toggle()
        self._app.emit("widget-layout-changed", self)

    def _show(self, toggled=False):
        """Show thumbnails when called from toggle.

        Args:
            toggled: If True thumbnail mode is already toggled.
        """
        # Clean liststore
        self._liststore.clear()

        # Draw the icon view instead of the image
        if not toggled:
            self._app["main_window"].switch_to_child(self)
        # Show the window
        super(Thumbnail, self).show()
        self.toggled = True

        # Add initial placeholder for all thumbnails
        default_pixbuf = self._get_default_pixbuf()
        for path in self._app.get_paths():
            name = self._get_name(path)
            self._liststore.append([default_pixbuf, name])

        # Generate thumbnails asynchronously
        self.reload_all(ignore_cache=True)

        # Set columns
        self.calculate_columns()

        # Focus the current image
        self.grab_focus()
        pos = self._app.get_index()
        self.move_to_pos(pos)

    def calculate_columns(self):
        """Calculate how many columns fit into the current window."""
        width = self._app["window"].winsize[0]
        if self._app["library"].grid.is_visible():
            width -= self._app["library"].get_size_request()[0]

        padding = self.get_item_padding()
        columns = floor((width - 12) / (self.get_zoom_level()[0] + 2 * padding))
        columns = max(1, columns)  # We may receive zero before
        free_space = (width - 12) % (
            self.get_zoom_level()[0] + 2 * padding)
        padding = floor(free_space / columns)
        self.set_column_spacing(padding)
        self.set_columns(columns)

    def _get_default_pixbuf(self):
        default_pixbuf_max = GdkPixbuf.Pixbuf.new_from_file_at_scale(
            self._thumbnail_manager.default_icon, *self.get_zoom_level(), True)
        size = self.get_zoom_level()[0]
        return self._thumbnail_manager.scale_pixbuf(default_pixbuf_max, size)

    def reload_all(self, ignore_cache=False):
        size = self.get_zoom_level()[0]
        for i, path in enumerate(self._app.get_paths()):
            self._thumbnail_manager.get_thumbnail_at_scale_async(
                path, size, self._on_thumbnail_created, i,
                ignore_cache=ignore_cache)

    def _on_thumbnail_created(self, pixbuf, position):
        # Happens if files are deleted while we are trying to create thumbnails
        # for them
        if len(self._liststore) > position:
            # Subscripting the liststore directly works fine
            # pylint: disable=unsubscriptable-object
            self._liststore[position][0] = pixbuf
            self.move_to_pos(self.get_position())

    def _get_name(self, filename):
        name = os.path.splitext(os.path.basename(filename))[0]
        if filename in self._app["mark"].marked:
            name += " [*]"

        return name

    def reload(self, filename, reload_image=True):
        """Reload the thumbnails of manipulated images.

        Args:
            filename: Name of the file to reload thumbnail of.
            reload_image: If True reload the image of the thumbnail. Else only
                the name (useful for marking).
        """
        index = self._app.get_paths().index(filename)
        name = self._get_name(filename)
        if os.path.basename(filename) \
                in self._app["commandline"].search.results:
            name = self._markup + "<b>" + name + "</b></span>"

        # pylint: disable=unsubscriptable-object
        if reload_image:
            self._thumbnail_manager.get_thumbnail_at_scale_async(
                filename, self.get_zoom_level()[0],
                self._on_thumbnail_created, index, ignore_cache=True)

        self._liststore[index][1] = name

    def move_direction(self, direction):
        """Scroll with "hjkl".

        Args:
            direction: Direction to scroll in. One of "hjkl".
        """
        # Start at current position
        new_pos = self.get_position()
        # Check for a user prefixed step
        step = self._app["eventhandler"].num_receive()
        # Get variables used for calculation of limits
        last = len(self._app.get_paths())
        rows = self.get_item_row(Gtk.TreePath(last - 1))
        columns = self.get_columns()
        elem_last_row = last - rows * columns
        elem_per_row = floor((last - elem_last_row) / rows) if rows else last
        column = self.get_item_column(Gtk.TreePath(new_pos))
        row = self.get_item_row(Gtk.TreePath(new_pos))
        min_pos = 0
        max_pos = last - 1
        # Simple scrolls
        if direction == "h":
            new_pos -= step
        elif direction == "k":
            min_pos = column
            new_pos -= columns * step
        elif direction == "l":
            new_pos += step
        elif direction == "j":
            max_pos = (rows - 1) * elem_per_row + column \
                if column >= elem_last_row else rows * elem_per_row + column
            new_pos += columns * step
        # First element in row
        elif direction == "H":
            new_pos = row * elem_per_row
        # Last element in column
        elif direction == "J":
            new_pos = (rows - 1) * elem_per_row + column \
                if column >= elem_last_row else rows * elem_per_row + column
        # First element in column
        elif direction == "K":
            new_pos %= elem_per_row
        # Last element in row
        elif direction == "L":
            new_pos = (row + 1) * elem_per_row - 1
        # Do not scroll to paths that are over the limits
        if new_pos < min_pos:
            new_pos = min_pos
        elif new_pos > max_pos:
            new_pos = max_pos
        # Move
        self.move_to_pos(new_pos)

    def move_to_pos(self, pos):
        """Set focus on position in iconview and center it.

        Args:
            pos: The position to focus.
        """
        self.select_path(Gtk.TreePath(pos))
        cell_renderer = self.get_cells()[0]
        self.set_cursor(Gtk.TreePath(pos), cell_renderer, False)
        self.scroll_to_path(Gtk.TreePath(pos), True, 0.5, 0.5)
        # Clear the user prefixed step
        self._app["eventhandler"].num_clear()

    def zoom(self, inc=True):
        """Zoom thumbnails.

        Args:
            inc: If True increase thumbnail size.
        """
        # What zoom and limits
        if inc and self._zoom_level_index < len(self._zoom_levels) - 1:
            self._zoom_level_index += 1
        elif not inc and self._zoom_level_index > 0:
            self._zoom_level_index -= 1
        else:
            return
        # Rescale all images in liststore
        if self.toggled:
            self.reload_all()
        # Set columns and refocus current image
        self.calculate_columns()
        self.move_to_pos(self.get_position())

    def get_zoom_level(self):
        return self._zoom_levels[self._zoom_level_index]

    def get_cache_directory(self):
        return self._thumbnail_manager.thumbnail_store.base_dir

    def get_position(self):
        path = self.get_cursor()[1]
        return path.get_indices()[0] if path else 0

    def on_paths_changed(self):
        """Reload thumbnails properly when paths have changed."""
        diff = len(self._liststore) - len(self._app.get_paths())
        # Delete extra elements
        while diff > 0:
            # pylint: disable=unsubscriptable-object
            self._liststore.remove(self._liststore[-1].iter)
            diff -= 1
        # Insert dummy elements to override
        default_pixbuf = self._get_default_pixbuf() if diff < 0 else None
        while diff < 0:
            name = self._get_name(self._app.get_paths()[diff])
            self._liststore.append([default_pixbuf, name])
            diff += 1
        for path in self._app.get_paths():
            self.reload(path)

    def _on_marks_changed(self, mark, changed):
        """Reload names if marks changed."""
        if self.toggled:
            for name in changed:
                self.reload(name, False)
        self._app["statusbar"].update_info()  # Do this once from here

    def _on_transformations_applied_to_file(self, transform, files):
        if self.toggled:
            for name in files:
                self.reload(name)

    def _on_search_completed(self, search, new_pos, last_focused):
        if self.toggled:
            for path in self._app.get_paths():
                self.reload(path, False)
        if last_focused == "thu":
            self.move_to_pos(new_pos)



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Class to handle log file for vimiv."""

import os
import sys
import time

from gi.repository import GLib, Gtk


class Log():
    """Log file handler.

    Attributes:
        _filename: Name of the log file.
        _terminal: sys.stderr to interact with
    """

    def __init__(self, app):
        """Create the necessary objects and settings.

        Args:
            app: The main vimiv application to interact with.
        """
        datadir = os.path.join(GLib.get_user_data_dir(), "vimiv")
        os.makedirs(datadir, exist_ok=True)
        self._filename = os.path.join(datadir, "vimiv.log")
        self._terminal = sys.stderr
        # Redirect stderr in debug mode so it is written to the log file as well
        if app.debug:
            sys.stderr = self
        # Create a new log file at startup
        with open(self._filename, "w") as f:
            f.write("Vimiv log written to "
                    + self._filename.replace(os.getenv("HOME"), "~")
                    + "\n")
        self._write_separator()
        # Header containing version and Gtk version
        self.write_message("Version", app["information"].get_version())
        self.write_message("Python", sys.version.split()[0])
        gtk_version = str(Gtk.get_major_version()) + "." \
            + str(Gtk.get_minor_version()) + "." \
            + str(Gtk.get_micro_version())
        self.write_message("GTK", gtk_version)
        self._write_separator()
        # Start time
        self.write_message("Started", "time")

    def write_message(self, header, message):
        """Write information to the log file.

        Args:
            header: Title of the message, gets surrounded in [].
            message: Log message.
        """
        if "time" in message:
            now = [str(t) for t in time.localtime()[3:6]]
            formatted_time = ":".join(now)
            message = message.replace("time", formatted_time)
        message = message.replace("\n", "\n" + " " * 16)
        formatted_message = "%-15s %s\n" % ("[" + header + "]", message)
        with open(self._filename, "a") as f:
            f.write(formatted_message)

    def write(self, message):
        """Write stderr message to log file and terminal."""
        print(message, end="")
        if "Traceback" in message:
            self.write_message("stderr", "")
        with open(self._filename, "a") as f:
            f.write(message)

    def flush(self):
        self._terminal.flush()

    def fileno(self):
        return self._terminal.fileno()

    def _write_separator(self):
        """Write a neat 80 * # separator to the log file."""
        with open(self._filename, "a") as f:
            f.write("#" * 80 + "\n")



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Actions which act on the actual image file."""

import os
from multiprocessing.pool import ThreadPool as Pool

from gi.repository import GdkPixbuf, GObject
from vimiv.fileactions import edit_supported

# We need the try ... except wrapper here
# pylint: disable=ungrouped-imports
try:
    from gi.repository import GExiv2
    _has_exif = True
except ImportError:
    _has_exif = False


def save_pixbuf(pixbuf, filename, update_orientation_tag=False):
    """Save the image with all the exif keys that exist if we have exif support.

    NOTE: This is used to override edited images, not to save images to new
        paths. The filename must exist as it is used to retrieve the image
        format and exif data.

    Args:
        pixbuf: GdkPixbuf.Pixbuf image to act on.
        filename: Name of the image to save.
        update_orientation_tag: If True, set orientation tag to NORMAL.
    """
    if not os.path.isfile(filename):
        raise FileNotFoundError("Original file to retrieve data from not found")
    # Get needed information
    info = GdkPixbuf.Pixbuf.get_file_info(filename)[0]
    extension = info.get_extensions()[0]
    if _has_exif:
        exif = GExiv2.Metadata(filename)
    # Save
    pixbuf.savev(filename, extension, [], [])
    if _has_exif and exif.get_supports_exif():
        if update_orientation_tag:
            exif.set_orientation(GExiv2.Orientation.NORMAL)
        exif.save_file()


def rotate_file(filename, cwise):
    """Rotate a file and save it.

    Args:
        filename: Name of the image to rotate.
        cwise: Rotate image 90 * cwise degrees.
    """
    pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)
    pixbuf = pixbuf.rotate_simple(90 * cwise)
    save_pixbuf(pixbuf, filename, update_orientation_tag=True)


def flip_file(filename, horizontal):
    """Flip a file and save it.

    Args:
        filename: Name of the image to flip.
        horizontal: If True, flip horizontally. Else vertically.
    """
    pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)
    pixbuf = pixbuf.flip(horizontal)
    save_pixbuf(pixbuf, filename)


class Autorotate(GObject.Object):
    """Class to rotate a list of images according to EXIF in a thread pool.

    Attributes:
        _filelist: List of files to rotate.
        _rotated_count: Int to count the amount of files that have been rotated.
        _processed_count: Int to count the amount of files that have been
            processed.
        _thread_pool: ThreadPool to use when rotating all images.

    Signals:
        completed: Emitted when all files where rotated so the statusbar can
            update.
    """

    def __init__(self, filelist):
        super(Autorotate, self).__init__()
        self._filelist = filelist
        self._rotated_count = 0
        self._processed_count = 0

        _cpu_count = os.cpu_count()
        if _cpu_count is None:
            _cpu_count = 1
        elif _cpu_count > 1:
            _cpu_count -= 1
        self._thread_pool = Pool(_cpu_count)

    def run(self):
        """Start autorotating the images in self._filelist."""
        for filename in self._filelist:
            self._thread_pool.apply_async(self._rotate, (filename,),
                                          callback=self._on_rotated)

    def _rotate(self, filename):
        """Rotate filename using pixbuf.apply_embedded_orientation()."""
        if not edit_supported(filename):
            return
        exif = GExiv2.Metadata(filename)
        if not exif.get_supports_exif():
            return
        orientation = exif.get_orientation()
        if orientation not in [GExiv2.Orientation.NORMAL,
                               GExiv2.Orientation.UNSPECIFIED]:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)
            pixbuf = pixbuf.apply_embedded_orientation()
            save_pixbuf(pixbuf, filename, update_orientation_tag=True)
            self._rotated_count += 1

    def _on_rotated(self, thread_pool_result):
        self._processed_count += 1
        if self._processed_count == len(self._filelist):
            self.emit("completed", self._rotated_count)


GObject.signal_new("completed", Autorotate, GObject.SIGNAL_RUN_LAST, None,
                   (GObject.TYPE_PYOBJECT,))



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Library part of vimiv."""

import os

from gi.repository import Gdk, Gtk
from vimiv.fileactions import is_image
from vimiv.helpers import listdir_wrapper, sizeof_fmt
from vimiv.settings import settings


class Library(Gtk.TreeView):
    """Library of vimiv.

    Includes the treeview with the library and all actions that apply to it.

    Attributes:
        files: Files in the library.
        grid: Gtk.Grid containing the TreeView and the border.

        _app: The main vimiv application to interact with.
        _positions: Dictionary that stores position in directories.
    """

    def __init__(self, app):
        """Create the necessary objects and settings.

        Args:
            app: The main vimiv application to interact with.
        """
        super(Library, self).__init__()
        self._app = app

        # Settings
        self._positions = {}
        border_width = settings["border_width"].get_value()

        # Defaults
        self.files = []

        # Grid with treeview and border
        self.grid = Gtk.Grid()
        # A simple border
        if border_width:
            border = Gtk.Separator()
            border.set_size_request(border_width, 1)
            self.grid.attach(border, 1, 0, 1, 1)
        # Pack everything
        self.set_size_request(
            settings["library_width"].get_value() - border_width, 10)
        scrolled_win = Gtk.ScrolledWindow()
        scrolled_win.set_vexpand(True)
        scrolled_win.add(self)
        scrolled_win.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.grid.attach(scrolled_win, 0, 0, 1, 1)
        # Treeview
        self.set_enable_search(False)
        # Select file when row activated
        self.connect("row-activated", self.file_select, True)
        # Handle key events
        self.add_events(Gdk.EventMask.KEY_PRESS_MASK)
        self.connect("key_press_event",
                     self._app["eventhandler"].on_key_press, "LIBRARY")
        self.connect("button_press_event",
                     self._app["eventhandler"].on_click, "LIBRARY")
        # Add the columns
        for i, name in enumerate(["Num", "Name", "Size", "M"]):
            renderer = Gtk.CellRendererText()
            column = Gtk.TreeViewColumn(name, renderer, markup=i)
            if name == "Name":
                column.set_expand(True)
                column.set_max_width(20)
            self.append_column(column)
        # Set the liststore model
        self.set_model(self._liststore_create())
        # Set the hexpand property if requested in the configfile
        if not self._app.get_paths() and settings["expand_lib"].get_value():
            self.set_hexpand(True)

        # Connect signals
        self._app.connect("paths-changed", self._on_paths_changed)
        self._app["mark"].connect("marks-changed", self._on_marks_changed)
        self._app["commandline"].search.connect("search-completed",
                                                self._on_search_completed)
        settings.connect("changed", self._on_settings_changed)

    def toggle(self, update_image=True):
        """Toggle the library.

        Args:
            update_image: If True update the image shown. Always the case except
            for when running toggle after file_select(), as file_select() does
            this by itself.
        """
        if self.grid.is_visible():
            self._remember_pos()
            self.grid.hide()
            self.focus(False)
        else:
            self.grid.show()
            if not self._app.get_paths():
                # Hide the non existing image and expand if necessary
                self._app["main_window"].hide()
                if settings["expand_lib"].get_value():
                    self.set_hexpand(True)
            else:  # Try to focus the current image in the library
                image = self._app.get_path()
                image_path = os.path.dirname(image)
                image_name = os.path.basename(image)
                if image_path == os.getcwd() and image_name in self.files:
                    self[os.getcwd()] = image_name
            # Stop the slideshow
            if self._app["slideshow"].running:
                self._app["slideshow"].toggle()
            self.focus(True)
            # Markings and other stuff might have changed
            self.reload(os.getcwd())
        self._app.emit("widget-layout-changed", self)

    def focus(self, focus_library=True):
        """Set or remove focus from the library.

        Args:
            focus_library: If True focus the library. Else unfocus it.
        """
        if focus_library:
            self.grab_focus()
            if not self.grid.is_visible():
                self.toggle()
        else:
            self._app["main_window"].grab_focus()
        # Update info for the current mode
        self._app["statusbar"].update_info()

    def file_select(self, treeview, path, column, close):
        """Show image or open directory for activated file in library.

        Args:
            treeview: The Gtk.TreeView which emitted the signal.
            path: Gtk.TreePath that was activated.
            column: Column that was activated.
            close: If True close the library when finished.
        """
        # Empty directory
        if not path:
            self._app["statusbar"].message("No file to select", "error")
            return
        count = path.get_indices()[0]
        fil = self.files[count]
        self[os.getcwd()] = fil
        if os.getcwd() == self._app["tags"].directory:
            self._tag_select(fil, close)
        elif os.path.isdir(fil):  # Open directory
            self.move_up(fil)
        else:
            self._image_select(fil, close)

    def _tag_select(self, tagname, close):
        # Also close if selected twice
        if tagname == self._app["tags"].last or close:
            self.toggle(update_image=False)
        self._app["tags"].load(tagname)

    def _image_select(self, basename, close):
        image = os.path.abspath(basename)
        # Close thumbnail
        if self._app["thumbnail"].toggled:
            self._app["thumbnail"].toggle()
            self.grab_focus()
        if self._app.get_paths() and image == self._app.get_path():
            close = True  # Close if file selected twice
        index = 0  # Catch directories to focus correctly
        for f in self.files:
            if f == basename:
                break
            elif os.path.isfile(f):
                index += 1
        # Repopulate
        visible_image = self._app.get_path() if self._app.get_paths() else ""
        self._app.populate([basename])
        if self._app.get_paths():
            self.set_hexpand(False)
            # Only load a new image if needed
            if visible_image != image:
                self._app["image"].move_index(
                    delta=index - self._app.get_index())
            # Close the library depending on key and repeat
            if close:
                self.toggle(update_image=False)
            self._app["main_window"].show()

    def move_up(self, directory="..", start=False):
        """Move up a directory or to a specific one in the library.

        Args:
            directory: Directory to move to. Defaults to parent.
            start: If True the function was called on startup and should not
                reload the library as it does not exist yet.
        """
        # Allow moving up multiple times if using .. as directory:
        if directory == "..":
            repeat = self._app["eventhandler"].num_receive()
            directory = "/".join(repeat * [".."])
        try:
            curdir = os.getcwd()
            os.chdir(directory)
            if not start:
                self.focus()
                self.reload(os.getcwd(), curdir)
        except (FileNotFoundError, PermissionError):
            self._app["statusbar"].message("Directory not accessible", "error")

    def reload(self, directory, last_directory="", search=False):
        """Reload the treeview.

        Args:
            directory: Directory of the library.
            last_directory: Directory that was last opened in the library.
            search: If True the reload request comes from a search
        """
        # Reset search positions
        if not search:
            self._app["commandline"].search.reset()
        # Create model in new directory
        self.set_model(self._liststore_create())
        # Warn if there are no files in the directory
        if not self.files:
            self._app["statusbar"].message("Directory is empty", "warning")
            return
        # Check if there is a saved position
        self.move_pos(True, self[directory])
        # Check if the last directory is in the current one
        if os.path.basename(last_directory) in self.files:
            self.move_pos(True,
                          self.files.index(os.path.basename(last_directory)))

    def reload_names(self):
        """Only reload names of the treeview."""
        model = self.get_model()
        for i, name in enumerate(self.files):
            markup_string = name
            if os.path.islink(name):
                markup_string += "  →  " + os.path.realpath(name)
            if os.path.isdir(name):
                markup_string = "<b>" + markup_string + "</b>"
            if name in self._app["commandline"].search.results:
                # This is a MarkupSetting not a BoolSetting as pylint thinks
                # pylint: disable=no-member
                markup_string = settings["markup"].surround(markup_string)
            model[i][1] = markup_string

    def move_pos(self, forward=True, defined_pos=None):
        """Move to a specific position in the library.

        Defaults to moving to the last file. Can be used for the first file or
        any defined position.

        Args:
            forward: If True move forwards.
            defined_pos: If not empty defines the position to move to.
        """
        if not self.files:
            self._app["statusbar"].message("No position to go to", "error")
            return
        max_pos = len(self.files)
        # Direct call from scroll
        if isinstance(defined_pos, int):
            new_pos = defined_pos
        elif forward:
            new_pos = self._app["eventhandler"].num_receive(max_pos) - 1
        else:
            new_pos = self._app["eventhandler"].num_receive() - 1
        if new_pos < 0 or new_pos > max_pos:
            self._app["statusbar"].message("Unsupported index", "warning")
            return
        self.set_cursor(Gtk.TreePath(new_pos), None, False)
        self.scroll_to_cell(Gtk.TreePath(new_pos), None, True, 0.5, 0)
        # Clear the prefix
        self._app["eventhandler"].num_clear()

    def scroll(self, direction):
        """Scroll the library viewer and call file_select if necessary.

        Args:
            direction: One of 'hjkl' defining the scroll direction.

        Return:
            True to deactivate default key-bindings for arrow keys.
        """
        # Handle the specific keys
        if direction == "h":  # Behave like ranger
            self._remember_pos()
            self.move_up()
        elif direction == "l":
            self.file_select(self, self.get_cursor()[0],
                             None, False)
        elif direction in ["j", "k"]:
            # Scroll the tree checking for a user step
            step = self._app["eventhandler"].num_receive()
            if direction == "j":
                new_pos = self.get_position() + step
                if new_pos >= len(self.get_model()):
                    new_pos = len(self.get_model()) - 1
            else:
                new_pos = self.get_position() - step
                if new_pos < 0:
                    new_pos = 0
            self.move_pos(True, new_pos)
        else:
            self._app["statusbar"].message(
                "Invalid scroll direction " + direction, "error")
        return True  # Deactivates default bindings (here for Arrows)

    def get_position(self):
        """Return focused position as integer."""
        path = self.get_cursor()[0]
        return path.get_indices()[0] if path else 0

    def _liststore_create(self):
        """Create the Gtk.ListStore containing information on supported files.

        Return:
            The created liststore containing
            [count, filename, filesize, markup_string].
        """
        liststore = Gtk.ListStore(int, str, str, str)
        self.files, filesize = self._filelist_create()
        # Remove unsupported files if one isn't in the tags directory
        if os.getcwd() != self._app["tags"].directory:
            self.files = [
                possible_file
                for possible_file in self.files
                if is_image(possible_file) or os.path.isdir(possible_file)]
        # Add all supported files
        for i, fil in enumerate(self.files):
            markup_string = fil
            size = filesize[fil]
            marked_string = ""
            if os.path.islink(fil):
                markup_string += "  →  " + os.path.realpath(fil)
            if os.path.abspath(fil) in self._app["mark"].marked:
                marked_string = "[*]"
            if os.path.isdir(fil):
                markup_string = "<b>" + markup_string + "</b>"
            if fil in self._app["commandline"].search.results:
                # This is a MarkupSetting not a BoolSetting as pylint thinks
                # pylint: disable=no-member
                markup_string = settings["markup"].surround(markup_string)
            liststore.append([i + 1, markup_string, size, marked_string])

        return liststore

    def _filelist_create(self, directory="."):
        """Create a filelist from all files in directory.

        Args:
            directory: Directory of which the filelist is created.
        Return:
            filelist, filesize: List of files, dictionary with filesize info
        """
        # Get data from ls -lh and parse it correctly
        files = listdir_wrapper(directory, settings["show_hidden"].get_value())
        filesize = {}
        file_check_amount = settings["file_check_amount"].get_value()
        for fil in files:
            # Catch broken symbolic links
            if os.path.islink(fil) and \
                    not os.path.exists(os.path.realpath(fil)):
                continue
            # Number of images in directory as filesize
            if os.path.isdir(fil):
                try:
                    subfiles = listdir_wrapper(
                        fil, settings["show_hidden"].get_value())
                    # Necessary to keep acceptable speed in library
                    many = False
                    if len(subfiles) > file_check_amount:
                        many = True
                    subfiles = [sub
                                for sub in subfiles[:file_check_amount]
                                if is_image(os.path.join(fil, sub))]
                    amount = str(len(subfiles))
                    if subfiles and many:
                        amount += "+"
                    filesize[fil] = amount
                except PermissionError:
                    filesize[fil] = "N/A"
            else:
                filesize[fil] = sizeof_fmt(os.path.getsize(fil))

        return files, filesize

    def _remember_pos(self):
        if self.files:
            self[os.getcwd()] = self.files[self.get_position()]

    def _on_paths_changed(self, app, widget):
        """Reload filelist on the paths-changed signal from app."""
        # Expand library if set by user and all paths were removed
        if not self._app.get_paths() and settings["expand_lib"].get_value():
            self.set_hexpand(True)
            if not self.is_focus():
                self.focus()
        if self.grid.is_visible():
            # Reload remembering path or staying as close as possible
            decremented_index = max(0, self.get_position() - 1)
            filename = self.files[self.get_position()]
            self.reload(os.getcwd())
            if filename in self.files:
                index = self.files.index(filename)
            else:
                index = min(decremented_index, len(self.files) - 1)
            self.move_pos(defined_pos=index)

    def _on_marks_changed(self, mark, changed):
        """Reload names if marks changed."""
        if self.grid.is_visible():
            model = self.get_model()
            for i, name in enumerate(self.files):
                model[i][3] = "[*]" \
                    if os.path.abspath(name) in mark.marked else ""

    def _on_search_completed(self, search, new_pos, last_focused):
        self.reload_names()
        # Move to next result
        if last_focused == "lib":
            self.move_pos(defined_pos=new_pos)
            if len(search.results) == 1 \
                    and not self._app["commandline"].is_visible():
                path = Gtk.TreePath(new_pos)
                self.file_select(self, path, None, False)

    def _on_settings_changed(self, new_settings, setting):
        if setting == "library_width":
            width = settings["library_width"].get_value()
            # Set some reasonable limits to the library size
            width = min(width, self._app["window"].winsize[0])
            width = max(width, 100)
            if width != settings["library_width"].get_value():
                settings.override("library_width", str(width))
                return
            else:
                self.set_size_request(width, 10)
                self._app.emit("widget-layout-changed", self)
        elif setting == "show_hidden" and self.is_visible():
            self.reload(".")

    def __getitem__(self, directory):
        """Convenience method to access saved positions via self[directory].

        Args:
            directory: Name of the directory to search for a saved position.
        Return:
            The index of the saved file if any.
        """
        if directory in self._positions:
            filename = self._positions[directory]
            if filename in self.files:
                return self.files.index(filename)
        return 0

    def __setitem__(self, directory, filename):
        """Convenience method to save positions in via self[directory].

        Args:
            directory: Name of the directory to search for a saved position.
            filename: Name of the file to save.
        """
        self._positions[directory] = filename



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Wrappers around standard library functions used in vimiv."""

import gzip
import os
import re

from gi.repository import Gtk
from vimiv.exceptions import StringConversionError


def listdir_wrapper(path, show_hidden=False):
    """Re-implementation of os.listdir which mustn't show hidden files.

    Args:
        path: Path of the directory in which os.listdir is called.
        show_hidden: If true, show hidden files. Else do not.
    Return:
        Sorted list of files in path.
    """
    all_files = sorted(os.listdir(os.path.expanduser(path)))
    if show_hidden:
        return all_files
    return [fil for fil in all_files if not fil.startswith(".")]


def read_file(filename):
    """Read the content of a file into a list or create file.

    Args:
        filename: The name of the file to be read.
    Return:
        Content of filename in a list.
    """
    content = []
    try:
        with open(filename, "r") as f:
            for line in f:
                content.append(line.rstrip("\n"))
    except FileNotFoundError:
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, "w") as f:
            f.write("")
    return content


def sizeof_fmt(num):
    """Print size of a byte number in human-readable format.

    Args:
        num: Filesize in bytes.

    Return:
        Filesize in human-readable format.
    """
    for unit in ["B", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1024.0:
            if abs(num) < 100:
                return "%3.1f%s" % (num, unit)
            return "%3.0f%s" % (num, unit)
        num /= 1024.0
    return "%.1f%s" % (num, "Y")


def error_message(message, running_tests=False):
    """Show a GTK Error Pop Up with message.

    Args:
        message: The message to display.
        running_tests: If True running from testsuite. Do not show popup.
    """
    # Always print the error message first
    print("\033[91mError:\033[0m", message)
    # Then display a Gtk Popup
    popup = Gtk.Dialog(title="vimiv - Error", transient_for=Gtk.Window())
    popup.set_default_size(600, 1)
    popup.add_button("Close", Gtk.ResponseType.CLOSE)
    message_label = Gtk.Label()
    # Set up label so it actually follows the default width of 600
    message_label.set_hexpand(True)
    message_label.set_line_wrap(True)
    message_label.set_size_request(600, 1)
    message_label.set_text(message)
    box = popup.get_child()
    box.set_border_width(12)
    grid = Gtk.Grid()
    grid.set_column_spacing(12)
    box.pack_start(grid, False, False, 0)
    icon_size = Gtk.IconSize(5)
    error_icon = Gtk.Image.new_from_icon_name("dialog-error", icon_size)
    grid.attach(error_icon, 0, 0, 1, 1)
    grid.attach(message_label, 1, 0, 1, 1)
    popup.show_all()
    if not running_tests:
        popup.run()
        popup.destroy()


def read_info_from_man():
    """Read information on commands from the vimivrc.5 man page.

    Return:
        A dictionary containing command-names and information on them.
    """
    infodict = {}
    # Man page might have been installed as .5.gz by a package manager
    if os.path.isfile("/usr/share/man/man5/vimivrc.5"):
        with open("/usr/share/man/man5/vimivrc.5") as f:
            man_text = f.read()
    elif os.path.isfile("/usr/share/man/man5/vimivrc.5.gz"):
        with gzip.open("/usr/share/man/man5/vimivrc.5.gz") as f:
            man_text = f.read().decode()
    else:
        return {}
    # Loop over lines receiving command name and info
    lines = man_text.split("\n")
    for i, line in enumerate(man_text.split("\n")):
        if line.startswith("\\fB\\fC"):
            name = line.replace("\\fB\\fC", "").replace("\\fR", "")
            info = lines[i + 1].rstrip(".")  # Remove trailing periods
            infodict[name] = info
    return infodict


def expand_filenames(filename, filelist, command):
    """Expand % to filename and * to filelist in command."""
    # Escape spaces for the shell
    filename = filename.replace(" ", "\\\\\\\\ ")
    filelist = [f.replace(" ", "\\\\\\\\ ") for f in filelist]
    # Substitute % and * with escaping
    command = re.sub(r'(?<!\\)(%)', filename, command)
    command = re.sub(r'(?<!\\)(\*)', " ".join(filelist), command)
    command = re.sub(r'(\\)(?!\\)', "", command)
    return command


def get_boolean(value):
    """Convert a value to a boolean.

    Args:
        value: String value to convert.
    Return:
        bool: True if string.lower() in ["yes", "true"]. False otherwise.
    """
    return True if value.lower() in ["yes", "true"] else False


def get_int(value, allow_sign=False):
    """Convert a value to an integer.

    Args:
        value: String value to convert.
        allow_sign: If True, negative values are allowed.
    Return:
        int(value) if possible.
    """
    try:
        # rstrip needed when 0. is passed via [count]
        int_val = int(value.rstrip("."))
    except ValueError:
        error = "Could not convert '%s' to int" % (value)
        raise StringConversionError(error)
    if int_val < 0 and not allow_sign:
        raise StringConversionError("Negative numbers are not supported.")
    return int_val


def get_float(value, allow_sign=False):
    """Convert a value to a float.

    Args:
        value: String value to convert.
        allow_sign: If True, negative values are allowed.
    Return:
        float(value) if possible.
    """
    try:
        float_val = float(value)
    except ValueError:
        error = "Could not convert '%s' to float" % (value)
        raise StringConversionError(error)
    if float_val < 0 and not allow_sign:
        raise StringConversionError("Negative numbers are not supported.")
    return float_val
